;; Generated package description from flymake-harper.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "flymake-harper" "0.1.0.0.20260820.5" "Flymake backend for Harper" '((emacs "27.1")) :commit "9327ba7fc17c7a3ef738c9be520c5ffad7a6d296" :authors '(("Manuel Uberti" . "manuel.uberti@inventati.org") ("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Philip Kaludercic" . "philipk@posteo.net") :keywords '("convenience" "text") :url "https://codeberg.org/pkal/flymake-harper.el")
